"""`libscarab` parameters taken directly from `parameters.h`

These need to be changed appropriately if corresponding values in
`parameters.h` are changed
"""

N      = 8
MU     = 4
NU     = 2**256
LOG_NU = 384
S1     = 8
S2     = 5
T      = 5            # Ceiling[Log[2, S2]] + 2
S      = 3            # Floor[Log[2, s2]] + 1


